!!!!!
!!!!! METHGI --- a program for maximization 
!!!!! Copyright 1984, 1985, 1989 A.M. Fedotov, E.Kh. Ginsburg and 
!!!!! T.I. Axenovich (Computing Center SD USSR and
!!!!! Institute of Cytology and Genetics SD USSR)
!!!!! Translation to Fortran-95 by Y. S. Aulchenko (ICG SD RAS), 2000
!!!!!

! 
! Main module
!
module methgimod

!maximal number of parameters under maximization
 integer, parameter 		:: MAXMPARAM=30	
!maximum number of iteration to do	
 integer, parameter 		:: NI=500
 integer, parameter 		:: IP=1
 double precision, parameter	:: PER=0.0001
 double precision, parameter	:: RED=0.5
 double precision, parameter	:: NEGLILH=-1.e18
!"small value"
 double precision, parameter	:: EPS=1.e-17

 type meth_param_type
!parameter value
  double precision	:: pr
!step (a value to add or subtract from parameter value while maximizing)
  double precision	:: st
!right boundary of parameter
  double precision	:: gr
!left boundary of parameter
  double precision	:: gl
!if parameter participates in maximization or is fixed
  integer		:: icp
!if parameter CURRENTLY participates in maximization or is fixed
  integer		:: isp
!if parameter participates in model (does not affect behavior of METHGI,
!use it for your own purposes)
  integer		:: ivec
!number of parameter... much other things about "inum"
  integer		:: inum
!parameter name
  character(20)		:: name
 end type meth_param_type

 type meth_type
  type(meth_param_type), dimension(MAXMPARAM)		:: param
!current value of the function under maximization
  double precision 					:: H
!number of iterations made
  integer						:: niter
!number of function calls
  integer						:: nfun
!number of marameters under maximization
  integer						:: nparam
 end type meth_type

contains

!
!load parameters from file "filename" and checks initial values
!
 function meth_loadparam(filename)
  character( * )	:: filename
  type(meth_type)	:: meth_loadparam
  character(1000)       :: rdline
  type(meth_param_type)	:: mde
  print *,'reading initial parameter state'
  open(11,file=filename,action='read',status='old',err=10)
  nparam=0
  do 
   read(11,'(a1000)',err=11,end=1) rdline
   if (rdline(1:1).ne.'#') then
    nparam=nparam+1
    read(rdline,*,err=12) mde%name,mde%inum,mde%ivec,mde%icp,mde%pr,&
&                  mde%st,mde%gl,mde%gr
    mde%isp=0
    print *,'reading and checking parameter ',nparam,' (',&
&mde%name(1:index(mde%name,' ')-1),')'
    if(mde%gl.ge.mde%gr) stop 'gl.ge.gr'
    if(mde%pr.lt.mde%gl) stop 'pr.lt.gl'
    if(mde%pr.gt.mde%gr) stop 'pr.gt.gr'
    if(mde%pr.eq.mde%gr.and.mde%icp.ne.0) stop 'pr.eq.gr.and.icp.ne.0'
    if(mde%pr.eq.mde%gl.and.mde%icp.ne.0) stop 'pr.eq.gl.and.icp.ne.0'
    if(abs(mde%st).lt.1.e-6) stop 'abs(st).lt.1.e-6'
    if(abs(mde%inum).gt.nparam) stop 'abs(inum).gt.nparam'
    meth_loadparam%param(nparam)=mde
   endif
  enddo
1 continue
  close(11)
  if(nparam.lt.1) stop 'number of parameters less then 1'
  meth_loadparam%nparam=nparam
  print *,'done. # parameters: ',meth_loadparam%nparam
  return
10 stop 'cannot open parameter file'
11 stop 'severe error while reading parameter file'
12 stop 'error while reading parameter file'
 end function meth_loadparam

!
!write maximized value of function and parameter estimates to "filename"
!
 subroutine meth_writeparam(filename,mdata)
  character(*),intent(IN) :: filename
  type(meth_type), intent(IN) :: mdata
  type(meth_param_type) :: mde
  print*,'writing output to ',filename
  open(11,file=filename,action='write',ERR=10)
   write(11,*)'number of iterations: ',mdata%niter
   write(11,*)'number of function calls: ',mdata%nfun
   write(11,*)'maximum function value: ',mdata%h
   write(11,*)'    PARAMETER ESTIMATES'
   do i=1,mdata%nparam
    mde=mdata%param(i)
    write(11,'(i2,1x,a10,3i2,2e14.5)',err=11)i,mde%name,mde%inum,mde%ivec,&
&                                           mde%icp,mde%pr,mde%st
   enddo
  close(11)
  print*,'written output'
  open(11,file=filename,action='write',ERR=10)
  return
10 stop 'cannt write output parameter file (1)'
11 stop 'cannt write output parameter file (2)'
 end subroutine meth_writeparam

!
! checks current parameter values. this function MUST be called before
! calculating each function value
!
 function   meth_checkparam(mdata)
  type(meth_type)	:: mdata
  double precision	:: meth_checkparam
  meth_checkparam=0.0
  np=mdata%nparam
  do i=1,np
   nnn=mdata%param(i)%inum
   if (nnn.ne.i) then
    if(nnn.gt.0) then
     mdata%param(i)%pr=mdata%param(nnn)%pr
    else
     nnn1=-1*nnn
     mdata%param(i)%pr=mdata%param(nnn1)%pr
    endif
   endif
  enddo
  do 2 i=1,np
   if(mdata%param(i)%icp.eq.0) goto 2
   if(mdata%param(i)%pr.le.mdata%param(i)%gl) goto 1
   if(mdata%param(i)%pr.ge.mdata%param(i)%gr) goto 1
2  continue
  return
1 meth_checkparam=NEGLILH
 end function   meth_checkparam

!
! Maximization procedure itself 
! It use obsolescent features of Fortran-77 19 times :)
!
	SUBROUTINE METHGI(mdata,FUN)
     	type(meth_type), intent(INOUT) :: mdata
        double precision, DIMENSION(maxmparam) :: Y
 	double precision :: begin,f,f1,x,z,ff
 	interface
 	 subroutine fun(LH)
	 double precision, intent(OUT) :: LH
	 end subroutine fun 
    	end interface
	
	nc=ni
	R=RED
	EP=EPS

        N=mdata%nparam
        JSP=0 
        IF (N)70,70,6 
6       IF (N-30)1,1,65
1       KK=0
        mdata%niter=0
        mdata%nfun=1
        BEGIN=-1.E18
        N1=0
        DO 2 J=1,N
        N1=N1+mdata%param(j)%icp
        IF (mdata%param(j)%icp.EQ.1)mdata%param(j)%isp=1
2       Y(J)=mdata%param(j)%pr 
        CALL FUN(F1)
        IF (F1+1.E17)75,75,101
101     DO 120 I=1,N
        mdata%param(i)%isp=0
120     CONTINUE
        DO 103 J=1,N
        IF (mdata%param(j)%icp.EQ.0)GOTO 103
        mdata%param(j)%isp=1
        IF (KK.EQ.1)mdata%param(j)%st=mdata%param(j)%ST*0.5 
20      K=2 
        mdata%param(j)%pr=mdata%param(j)%pr-mdata%param(j)%st
102     CALL FUN(F) 
        IF (F.LE.-1.E17)GOTO 107 
        IF (JSP.EQ.0)GOTO 107 
        mdata%param(jsp)%ISP=0
        JSP=0 
107     mdata%nfun=mdata%nfun+1 
        IF (F1-F)23,22,22 
22      K=K+1 
        mdata%param(j)%pr=mdata%param(j)%pr+2.*mdata%param(j)%st 
        IF (K-3)105,31,106
31      mdata%param(j)%st=-1.0*mdata%param(j)%ST*.5 
        GO TO 102 
23      mdata%param(j)%st=mdata%param(j)%ST*1.5 
        F1=F
105     mdata%param(j)%isp=0
        GOTO 103
106     JSP=J 
103     CONTINUE
        JSP=0 
        DO 24 J=1,N 
        IF (mdata%param(j)%icp.EQ.0)GOTO 24 
        X=2.*mdata%param(j)%pr-Y(J)
        Z=mdata%param(j)%pr-Y(J) 
        IF (Z)201,203,202 
201     IF (mdata%param(j)%GL-X)203,211,211 
211     X=mdata%param(j)%GL+eps
        GOTO 203
202     IF (mdata%param(j)%gr-X)212,212,203 
212     X=mdata%param(j)%gr-eps
203     CONTINUE
        Y(J)=mdata%param(j)%pr 
        mdata%param(j)%pr=X
        mdata%param(j)%isp=1
24      CONTINUE
        K=5 
104     CALL FUN(F) 
        mdata%nfun=mdata%nfun+1 
        IF (F-F1)25,25,50 
25      K=K+1 
        DO 26 J=1,N 
        IF (mdata%param(j)%icp.EQ.0)GOTO 26 
        mdata%param(j)%pr=Y(J)*(1.-R)+mdata%param(j)%pr*R 
26      CONTINUE
        IF (K-15)104,27,27
27      KK=1
        CALL FUN(F) 
        mdata%nfun=mdata%nfun+1 
        GO TO 53
50      KK=0
        IF (F-BEGIN-EP)521,53,53
521     CONTINUE
	mdata%h=f
	if(f.gt.-1.e18)goto 599
	mdata%h=f1
	do 600 i=1,n
600     mdata%param(i)%pr=y(i)
599     mdata%niter=mdata%NITER+1 
        GO TO 512 
53      KP=0
        DO 54 J=1,N 
        IF (mdata%param(j)%icp.EQ.0)GOTO 54 
        IF (ABS(PER*mdata%param(j)%pr)-ABS(mdata%param(j)%st))45,531,531
45      IF (ABS(mdata%param(j)%st)-1.E-06)531,54,54
531     KP=KP+1 
54      CONTINUE
        IF (KP-N1)52,521,521
52      mdata%niter=mdata%NITER+1 
        IF (mdata%niter-NC)51,51,511 
511     CONTINUE
512     CONTINUE
        RETURN

51	FF=F1
	IF (F.GT.FF) FF=F
5111        IF (IP)501,55,501
501	FF=F1
	IF (F.GT.FF) FF=F

55      IF (F+1.E17)551,551,550 
550     F1=F
551     BEGIN=F1
        GO TO 101 
65      continue
        NP=-NP
        RETURN
70      continue
        RETURN
75      continue
        NP=0
        RETURN
        END subroutine methgi

end module methgimod
